@@ -1,4 +1,4 @@
-/* Copyright (C) 1999 Free Software Foundation, Inc.
+/* Copyright (C) 1999-2001 Free Software Foundation, Inc.
    This file is part of the GNU UTF-8 Library.
 
    The GNU UTF-8 Library is free software; you can redistribute it and/or
@@ -32,7 +32,7 @@ wint_t towupper (wint_t wc)
     else
       return wc;
   } else {
-    if ((unsigned int) wc <= 0xffffu)
+    if ((unsigned int) wc < 0x10800u)
       return (unsigned short) (wc + toupper_table[((unsigned int) wc) >> 8][((unsigned int) wc) & 0xff]);
     else
       return wc;
