@@ -1,4 +1,4 @@
-/* Copyright (C) 1999 Free Software Foundation, Inc.
+/* Copyright (C) 1999-2001 Free Software Foundation, Inc.
    This file is part of the GNU UTF-8 Library.
 
    The GNU UTF-8 Library is free software; you can redistribute it and/or
@@ -41,7 +41,7 @@ struct unicode_attribute {
 
 #define NONE (~(unsigned int)0)
 
-struct unicode_attribute unicode_attributes [0x10000];
+struct unicode_attribute unicode_attributes [0x110000];
 
 static void fill_attribute (unsigned int i,
                             const char* field1, const char* field2,
@@ -54,10 +54,12 @@ static void fill_attribute (unsigned int i,
 {
   struct unicode_attribute * uni;
 
-  if (i >= 0x10000) {
+  if (i >= 0x110000) {
     fprintf(stderr, "index too large\n");
     exit(1);
   }
+  if (strcmp(field2,"Cs") == 0)
+    return;
   uni = &unicode_attributes[i];
   uni->name = strdup(field1);
   uni->category = (field2[0]=='\0' ? "" : strdup(field2));
@@ -106,7 +108,7 @@ static void fill_attributes (const char* unicodedata_filename)
   char field14 [100];
   int lineno = 0;
 
-  for (i = 0; i < 0x10000; i++)
+  for (i = 0; i < 0x110000; i++)
     unicode_attributes[i].name = NULL;
 
   f = fopen(unicodedata_filename, "r");
@@ -218,7 +220,7 @@ static int isCLbothcase (unsigned int ch)
 /* Create uni_upcase.c, used by clisp. */
 void output_clisp_upcase_table (void)
 {
-  int pages[0x100];
+  int pages[0x1100];
   int p, p1, p2, i1, i2;
   const char* filename = "uni_upcase.c";
   FILE* f = fopen(filename, "w");
@@ -233,9 +235,9 @@ void output_clisp_upcase_table (void)
   fprintf(f, " * Generated automatically by the gentables utility.\n");
   fprintf(f, " */\n");
   fprintf(f, "\n");
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     pages[p] = 0;
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     for (i1 = 0; i1 < 0x100; i1++) {
       unsigned int ch = 0x100*p + i1;
       if (isCLbothcase(ch) && (uppercase(ch) != ch)) {
@@ -243,7 +245,7 @@ void output_clisp_upcase_table (void)
         break;
       }
     }
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     if (pages[p]) {
       fprintf(f, "static const cint up_case_table_page%02x[256] = {\n", p);
       for (i1 = 0; i1 < 32; i1++) {
@@ -252,6 +254,10 @@ void output_clisp_upcase_table (void)
           unsigned int ch = 256*p + 8*i1 + i2;
           unsigned int ch2 = (isCLbothcase(ch) ? uppercase(ch) : ch);
           int j = ((int)ch2 - (int)ch) & 0xffff;
+          if (ch2 != ch + (short)j) {
+            fprintf(stderr, "upper case maps 0x%04x to different plane\n", ch);
+            exit(1);
+          }
           fprintf(f, "0x%04x%s ", j, (8*i1+i2<255?",":" "));
         }
         fprintf(f, "/* 0x%02x-0x%02x */\n", 8*i1, 8*i1+7);
@@ -259,15 +265,15 @@ void output_clisp_upcase_table (void)
       fprintf(f, "};\n");
       fprintf(f, "\n");
     }
-  fprintf(f, "static const cint * const up_case_table[256] = {\n");
-  for (p1 = 0; p1 < 64; p1++) {
+  fprintf(f, "static const cint * const up_case_table[0x1100] = {\n");
+  for (p1 = 0; p1 < 0x440; p1++) {
     fprintf(f, "  ");
     for (p2 = 0; p2 < 4; p2++) {
       p = 4*p1 + p2;
       if (pages[p])
-        fprintf(f, "up_case_table_page%02x%s ", p, (p<255?",":" "));
+        fprintf(f, "up_case_table_page%02x%s ", p, (p<0x1100-1?",":" "));
       else
-        fprintf(f, "nop_page%s ", (p<255?",":" "));
+        fprintf(f, "nop_page%s ", (p<0x1100-1?",":" "));
     }
     fprintf(f, "/* 0x%02x-0x%02x */\n", 4*p1, 4*p1+3);
   }
@@ -286,7 +292,7 @@ void output_clisp_upcase_table (void)
 /* Create uni_downcase.c, used by clisp. */
 void output_clisp_downcase_table (void)
 {
-  int pages[0x100];
+  int pages[0x1100];
   int p, p1, p2, i1, i2;
   const char* filename = "uni_downcase.c";
   FILE* f = fopen(filename, "w");
@@ -301,9 +307,9 @@ void output_clisp_downcase_table (void)
   fprintf(f, " * Generated automatically by the gentables utility.\n");
   fprintf(f, " */\n");
   fprintf(f, "\n");
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     pages[p] = 0;
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     for (i1 = 0; i1 < 0x100; i1++) {
       unsigned int ch = 0x100*p + i1;
       if (isCLbothcase(ch) && (lowercase(ch) != ch)) {
@@ -311,7 +317,7 @@ void output_clisp_downcase_table (void)
         break;
       }
     }
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     if (pages[p]) {
       fprintf(f, "static const cint down_case_table_page%02x[256] = {\n", p);
       for (i1 = 0; i1 < 32; i1++) {
@@ -320,6 +326,10 @@ void output_clisp_downcase_table (void)
           unsigned int ch = 256*p + 8*i1 + i2;
           unsigned int ch2 = (isCLbothcase(ch) ? lowercase(ch) : ch);
           int j = ((int)ch2 - (int)ch) & 0xffff;
+          if (ch2 != ch + (short)j) {
+            fprintf(stderr, "lower case maps 0x%04x to different plane\n", ch);
+            exit(1);
+          }
           fprintf(f, "0x%04x%s ", j, (8*i1+i2<255?",":" "));
         }
         fprintf(f, "/* 0x%02x-0x%02x */\n", 8*i1, 8*i1+7);
@@ -327,15 +337,15 @@ void output_clisp_downcase_table (void)
       fprintf(f, "};\n");
       fprintf(f, "\n");
     }
-  fprintf(f, "static const cint * const down_case_table[256] = {\n");
-  for (p1 = 0; p1 < 64; p1++) {
+  fprintf(f, "static const cint * const down_case_table[0x1100] = {\n");
+  for (p1 = 0; p1 < 0x440; p1++) {
     fprintf(f, "  ");
     for (p2 = 0; p2 < 4; p2++) {
       p = 4*p1 + p2;
       if (pages[p])
-        fprintf(f, "down_case_table_page%02x%s ", p, (p<255?",":" "));
+        fprintf(f, "down_case_table_page%02x%s ", p, (p<0x1100-1?",":" "));
       else
-        fprintf(f, "nop_page%s ", (p<255?",":" "));
+        fprintf(f, "nop_page%s ", (p<0x1100-1?",":" "));
     }
     fprintf(f, "/* 0x%02x-0x%02x */\n", 4*p1, 4*p1+3);
   }
@@ -353,7 +363,7 @@ void output_clisp_downcase_table (void)
 
 static void output_copyright (FILE* f)
 {
-  fprintf(f, "/* Copyright (C) 1999 Free Software Foundation, Inc.\n");
+  fprintf(f, "/* Copyright (C) 1999-2001 Free Software Foundation, Inc.\n");
   fprintf(f, "   This file is part of the GNU UTF-8 Library.\n");
   fprintf(f, "\n");
   fprintf(f, "   The GNU UTF-8 Library is free software; you can redistribute it and/or\n");
@@ -375,7 +385,7 @@ static void output_copyright (FILE* f)
 /* Create toupper.h, used by libutf8. */
 void output_toupper_table (void)
 {
-  int pages[0x100];
+  int pages[0x1100];
   int p, p1, p2, i1, i2;
   const char* filename = "toupper.h";
   FILE* f = fopen(filename, "w");
@@ -388,9 +398,9 @@ void output_toupper_table (void)
   fprintf(f, "/* toupper table */\n");
   fprintf(f, "/* Generated automatically by the gentables utility. */\n");
   fprintf(f, "\n");
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     pages[p] = 0;
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     for (i1 = 0; i1 < 0x100; i1++) {
       unsigned int ch = 0x100*p + i1;
       if (uppercase(ch) != ch) {
@@ -398,7 +408,7 @@ void output_toupper_table (void)
         break;
       }
     }
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     if (pages[p]) {
       fprintf(f, "static const short toupper_table_page%02x[256] = {\n", p);
       for (i1 = 0; i1 < 32; i1++) {
@@ -407,6 +417,10 @@ void output_toupper_table (void)
           unsigned int ch = 256*p + 8*i1 + i2;
           unsigned int ch2 = uppercase(ch);
           int j = ((int)ch2 - (int)ch) & 0xffff;
+          if (ch2 != ch + (short)j) {
+            fprintf(stderr, "toupper maps 0x%04x to different plane\n", ch);
+            exit(1);
+          }
           fprintf(f, "0x%04x%s ", j, (8*i1+i2<255?",":" "));
         }
         fprintf(f, "/* 0x%02x-0x%02x */\n", 8*i1, 8*i1+7);
@@ -414,15 +428,15 @@ void output_toupper_table (void)
       fprintf(f, "};\n");
       fprintf(f, "\n");
     }
-  fprintf(f, "const short * const toupper_table[256] = {\n");
-  for (p1 = 0; p1 < 64; p1++) {
+  fprintf(f, "const short * const toupper_table[0x1100] = {\n");
+  for (p1 = 0; p1 < 0x440; p1++) {
     fprintf(f, "  ");
     for (p2 = 0; p2 < 4; p2++) {
       p = 4*p1 + p2;
       if (pages[p])
-        fprintf(f, "toupper_table_page%02x%s ", p, (p<255?",":" "));
+        fprintf(f, "toupper_table_page%02x%s ", p, (p<0x1100-1?",":" "));
       else
-        fprintf(f, "nop_page%s ", (p<255?",":" "));
+        fprintf(f, "nop_page%s ", (p<0x1100-1?",":" "));
     }
     fprintf(f, "/* 0x%02x-0x%02x */\n", 4*p1, 4*p1+3);
   }
@@ -441,7 +455,7 @@ void output_toupper_table (void)
 /* Create tolower.h, used by libutf8. */
 void output_tolower_table (void)
 {
-  int pages[0x100];
+  int pages[0x1100];
   int p, p1, p2, i1, i2;
   const char* filename = "tolower.h";
   FILE* f = fopen(filename, "w");
@@ -454,9 +468,9 @@ void output_tolower_table (void)
   fprintf(f, "/* tolower table */\n");
   fprintf(f, "/* Generated automatically by the gentables utility. */\n");
   fprintf(f, "\n");
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     pages[p] = 0;
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     for (i1 = 0; i1 < 0x100; i1++) {
       unsigned int ch = 0x100*p + i1;
       if (lowercase(ch) != ch) {
@@ -464,7 +478,7 @@ void output_tolower_table (void)
         break;
       }
     }
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     if (pages[p]) {
       fprintf(f, "static const short tolower_table_page%02x[256] = {\n", p);
       for (i1 = 0; i1 < 32; i1++) {
@@ -473,6 +487,10 @@ void output_tolower_table (void)
           unsigned int ch = 256*p + 8*i1 + i2;
           unsigned int ch2 = lowercase(ch);
           int j = ((int)ch2 - (int)ch) & 0xffff;
+          if (ch2 != ch + (short)j) {
+            fprintf(stderr, "tolower maps 0x%04x to different plane\n", ch);
+            exit(1);
+          }
           fprintf(f, "0x%04x%s ", j, (8*i1+i2<255?",":" "));
         }
         fprintf(f, "/* 0x%02x-0x%02x */\n", 8*i1, 8*i1+7);
@@ -480,15 +498,15 @@ void output_tolower_table (void)
       fprintf(f, "};\n");
       fprintf(f, "\n");
     }
-  fprintf(f, "const short * const tolower_table[256] = {\n");
-  for (p1 = 0; p1 < 64; p1++) {
+  fprintf(f, "const short * const tolower_table[0x1100] = {\n");
+  for (p1 = 0; p1 < 0x440; p1++) {
     fprintf(f, "  ");
     for (p2 = 0; p2 < 4; p2++) {
       p = 4*p1 + p2;
       if (pages[p])
-        fprintf(f, "tolower_table_page%02x%s ", p, (p<255?",":" "));
+        fprintf(f, "tolower_table_page%02x%s ", p, (p<0x1100-1?",":" "));
       else
-        fprintf(f, "nop_page%s ", (p<255?",":" "));
+        fprintf(f, "nop_page%s ", (p<0x1100-1?",":" "));
     }
     fprintf(f, "/* 0x%02x-0x%02x */\n", 4*p1, 4*p1+3);
   }
@@ -520,8 +538,8 @@ void output_tolower_table (void)
 /* Create attribute.h, used by libutf8. */
 void output_attribute_table (void)
 {
-  int table[0x10000];
-  int pages[0x100];
+  int table[0x110000];
+  int pages[0x1100];
   int p, q, p1, p2, i;
   unsigned int ch;
   const char* filename = "attribute.h";
@@ -535,7 +553,7 @@ void output_attribute_table (void)
   fprintf(f, "/* attribute table */\n");
   fprintf(f, "/* Generated automatically by the gentables utility. */\n");
   fprintf(f, "\n");
-  for (ch = 0; ch < 0x10000; ch++) {
+  for (ch = 0; ch < 0x110000; ch++) {
     int attributes = 0;
     if ((uppercase(ch) == ch || titlecase(uppercase(ch)) == ch)
         && lowercase(ch) != ch)
@@ -593,7 +611,7 @@ void output_attribute_table (void)
       attributes |= alnum;
     table[ch] = attributes;
   }
-  for (p = 0; p < 0x100; p++) {
+  for (p = 0; p < 0x1100; p++) {
     pages[p] = -1;
     for (q = 0; q < p; q++)
       if (pages[q] < 0) {
@@ -609,7 +627,7 @@ void output_attribute_table (void)
         }
       }
   }
-  for (p = 0; p < 0x100; p++)
+  for (p = 0; p < 0x1100; p++)
     if (pages[p] < 0) {
       fprintf(f, "static const unsigned char attribute_table_page%02x[256] = {\n", p);
       for (i = 0; i < 0x100; i++) {
@@ -686,14 +704,13 @@ void output_attribute_table (void)
       fprintf(f, "};\n");
       fprintf(f, "\n");
     }
-  fprintf(f, "const unsigned char * const attribute_table[256] = {\n");
-  for (p1 = 0; p1 < 128; p1++) {
+  fprintf(f, "const unsigned char * const attribute_table[0x1100] = {\n");
+  for (p1 = 0; p1 < 0x880; p1++) {
     fprintf(f, "  ");
     for (p2 = 0; p2 < 2; p2++) {
       p = 2*p1 + p2;
-      if (pages[p] >= 0)
-        p = pages[p];
-      fprintf(f, "attribute_table_page%02x%s ", p, (p<255?",":" "));
+      fprintf(f, "attribute_table_page%02x%s ", (pages[p] >= 0 ? pages[p] : p),
+                 (p<0x1100-1?",":" "));
     }
     fprintf(f, "/* 0x%02x-0x%02x */\n", 2*p1, 2*p1+1);
   }
@@ -718,20 +735,23 @@ static int nonspacing (unsigned int ch)
   const char* combining = unicode_attributes[ch].combining;
   if (combining != NULL && strcmp(combining,"") && strcmp(combining,"0"))
     return 1;
-  else
-    return 0;
 #else
   if (unicode_attributes[ch].bidi != NULL && !strcmp(unicode_attributes[ch].bidi,"NSM"))
     return 1;
-  else
-    return 0;
 #endif
+  /* Format control characters have width 0 as well. */
+  if (unicode_attributes[ch].bidi != NULL && !strcmp(unicode_attributes[ch].category,"Cf"))
+    return 1;
+  /* Zero width characters have width 0 as well. */
+  if (unicode_attributes[ch].name != NULL && !strncmp(unicode_attributes[ch].name,"ZERO WIDTH ",11))
+    return 1;
+  return 0;
 }
 
 /* Create nonspacing.h, used by libutf8. */
 void output_nonspacing_table (void)
 {
-  int pages[0x40];
+  int pages[0x440];
   int p, p1, p2, i1, i2, i3;
   const char* filename = "nonspacing.h";
   FILE* f = fopen(filename, "w");
@@ -744,9 +764,9 @@ void output_nonspacing_table (void)
   fprintf(f, "/* Non-spacing attribute table */\n");
   fprintf(f, "/* Generated automatically by the gentables utility. */\n");
   fprintf(f, "\n");
-  for (p = 0; p < 0x40; p++)
+  for (p = 0; p < 0x440; p++)
     pages[p] = 0;
-  for (p = 0; p < 0x40; p++)
+  for (p = 0; p < 0x440; p++)
     for (i1 = 0; i1 < 0x400; i1++) {
       unsigned int ch = 0x400*p + i1;
       if (nonspacing(ch)) {
@@ -754,7 +774,7 @@ void output_nonspacing_table (void)
         break;
       }
     }
-  for (p = 0; p < 0x40; p++)
+  for (p = 0; p < 0x440; p++)
     if (pages[p]) {
       fprintf(f, "static const unsigned char nonspacing_table_page%02x[128] = {\n", 4*p);
       for (i1 = 0; i1 < 16; i1++) {
@@ -773,15 +793,15 @@ void output_nonspacing_table (void)
       fprintf(f, "};\n");
       fprintf(f, "\n");
     }
-  fprintf(f, "static const unsigned char * const nonspacing_table[64] = {\n");
-  for (p1 = 0; p1 < 16; p1++) {
+  fprintf(f, "static const unsigned char * const nonspacing_table[0x440] = {\n");
+  for (p1 = 0; p1 < 0x110; p1++) {
     fprintf(f, "  ");
     for (p2 = 0; p2 < 4; p2++) {
       p = 4*p1 + p2;
       if (pages[p])
-        fprintf(f, "nonspacing_table_page%02x%s ", 4*p, (p<63?",":" "));
+        fprintf(f, "nonspacing_table_page%02x%s ", 4*p, (p<0x440-1?",":" "));
       else
-        fprintf(f, "NULL%s ", (p<63?",":" "));
+        fprintf(f, "NULL%s ", (p<0x440-1?",":" "));
     }
     fprintf(f, "/* 0x%04x-0x%04x */\n", 0x1000*p1, 0x1000*p1+0xfff);
   }
